--===========================================================================--
--
--  System Interface - For XSA-100 Spartan II BOARD
--
--  John Kent 14th July 2001
--  John Kent 1st April 2003 - Modified for Xess XSA-100
--  John Kent 19th May 2003 - Modified for wishbone miniUart
--
-- Uses rxunit.vhd, tx_unit.vhd, clkunit.vhd
-- from miniUart in the open cores site.
--
-------------------------------------------------------------------------------
library ieee;
   use ieee.std_logic_1164.all;
   use IEEE.STD_LOGIC_ARITH.ALL;
   use IEEE.STD_LOGIC_UNSIGNED.ALL;
   use ieee.numeric_std.all;

entity Micro8 is
  port(
    SysClk      : in  Std_Logic;  -- System Clock input
	 Reset_n     : in  Std_logic;  -- Master Reset input (active low)
    led         : out std_logic;  -- Diagnostic LED Flasher

    -- Memory Interface signals
    ram_csn     : out Std_Logic;
    ram_rdn     : out Std_Logic;
    ram_wrn     : out Std_Logic;
    ram_addr    : out Std_Logic_Vector(16 downto 0);
    ram_data    : inout Std_Logic_Vector(7 downto 0);

	 -- Uart Interface
    rxbit       : in  Std_Logic;
	 txbit       : out Std_Logic;

    flash_csn   : out Std_Logic
	 );
end Micro8;

-------------------------------------------------------------------------------
-- Architecture for memio Controller Unit
-------------------------------------------------------------------------------
architecture my_computer of Micro8 is
  -----------------------------------------------------------------------------
  -- Signals
  -----------------------------------------------------------------------------
  -- BOOT ROM
  signal rom_data_out  : Std_Logic_Vector(7 downto 0);
  signal rom_stb       : Std_Logic;
  signal rom_ack       : Std_Logic;

  -- UART Interface signals
  signal uart_data_out : Std_Logic_Vector(7 downto 0);  
  signal uart_stb      : Std_Logic;
  signal uart_ack      : Std_Logic;

  -- RAM
  signal ram_stb       : std_logic; -- memory chip select strobe
  signal ram_ack       : std_logic; -- memory chip select acknowledge
  signal ram_rd        : std_logic; -- memory read
  signal ram_wr        : std_logic; -- memory write
  signal ram_data_out  : std_logic_vector(7 downto 0);

  -- Sequencer Interface signals
  signal cpu_reset   : Std_Logic;
  signal cpu_we      : std_logic;
  signal cpu_stb     : std_logic;
  signal cpu_ack     : std_logic;
  signal cpu_addr    : Std_Logic_Vector(10 downto 0);
  signal cpu_data_in : Std_Logic_Vector(7 downto 0);
  signal cpu_data_out: Std_Logic_Vector(7 downto 0);
  signal cpu_irq_0   : std_logic;
  signal cpu_irq_1   : std_logic;

  -- Memory Transfer signals
  signal countL      : std_logic_vector(23 downto 0);

-----------------------------------------------------------------
--
-- Open Cores Mini UART
--
-----------------------------------------------------------------

component UART is
  generic(BRDIVISOR: INTEGER range 0 to 65535 := 130); -- baudrate divisor
  port (
-- Wishbone signals
     WB_CLK_I : in  std_logic;  -- clock
     WB_RST_I : in  std_logic;  -- Reset input
     WB_ADR_I : in  std_logic_vector(1 downto 0); -- Adress bus          
     WB_DAT_I : in  std_logic_vector(7 downto 0); -- DataIn Bus
     WB_DAT_O : out std_logic_vector(7 downto 0); -- DataOut Bus
     WB_WE_I  : in  std_logic;  -- Write Enable
     WB_STB_I : in  std_logic;  -- Strobe
     WB_ACK_O : out std_logic;	-- Acknowledge
-- process signals     
     IntTx_O  : out std_logic;  -- Transmit interrupt: indicate waiting for Byte
     IntRx_O  : out std_logic;  -- Receive interrupt: indicate Byte received
     BR_Clk_I : in  std_logic;  -- Clock used for Transmit/Receive
     TxD_PAD_O: out std_logic;  -- Tx RS232 Line
     RxD_PAD_I: in  std_logic);  -- Rx RS232 Line     
end component UART;

component CPU8BIT2 is
	port (	
     WB_CLK_I : in  std_logic;  -- clock
     WB_RST_I : in  std_logic;  -- Reset input
     WB_ADR_O : out std_logic_vector(10 downto 0); -- Adress bus          
     WB_DAT_I : in  std_logic_vector( 7 downto 0); -- DataIn Bus
     WB_DAT_O : out std_logic_vector( 7 downto 0); -- DataOut Bus
     WB_WE_O  : out std_logic;   -- Write Enable
     WB_STB_O : out std_logic;   -- Strobe
     WB_ACK_I : in  std_logic	   -- Acknowledge
);
end component CPU8BIT2;

component boot_rom is
  port (
    WB_ADR_I  : in  Std_Logic_Vector(5 downto 0);  -- 64 byte boot rom
	 WB_DAT_O  : out Std_Logic_Vector(7 downto 0);
	 WB_STB_I  : in  Std_Logic;
	 WB_ACK_O  : out Std_Logic
  );
end component boot_rom;

--component BUFG is 
--  port (
--     i: in std_logic;
--	    o: out std_logic
--  );
--end component;

begin
  -----------------------------------------------------------------------------
  -- Instantiation of internal components
  -----------------------------------------------------------------------------

my_uart  : UART
    generic map(BRDIVISOR => 260)     
    port map (
-- Wishbone signals
    WB_CLK_I  => SysClk,
	 WB_RST_I  => cpu_reset,
    WB_ADR_I  => cpu_addr(1 downto 0),
	 WB_DAT_I  => cpu_data_out,
	 WB_DAT_O  => uart_data_out,
	 WB_WE_I   => cpu_we,
    WB_STB_I  => uart_stb,
	 WB_ACK_O  => uart_ack,
-- process signals     
	 IntTx_O   => cpu_irq_1,
	 IntRx_O   => cpu_irq_0,
    BR_Clk_I  => SysClk,
	 TxD_PAD_O => txbit,
	 RxD_PAD_I => rxbit
	 );

my_cpu : CPU8BIT2 port map (    
     WB_CLK_I => SysClk,     -- clock
     WB_RST_I => cpu_reset,  -- Reset input
     WB_ADR_O => cpu_addr(10 downto 0),         
     WB_DAT_I => cpu_data_in,
     WB_DAT_O => cpu_data_out,
     WB_WE_O  => cpu_we,     -- Write Enable
     WB_STB_O => cpu_stb,    -- Strobe
     WB_ACK_I => cpu_ack	  -- Acknowledge
  );


my_rom : boot_rom port map (
	 WB_ADR_I   => cpu_addr(5 downto 0),
    WB_DAT_O   => rom_data_out,
	 WB_STB_I   => rom_stb,
	 WB_ACK_O   => rom_ack
	 );

-- bufginst: BUFG port map(i => countL(21), o => cpu_clk);
-- bufginst: BUFG port map(i => SysClk, o => cpu_clk );	 
	 
----------------------------------------------------------------------
--
--  Processes to decode memory devices
--
----------------------------------------------------------------------

decode: process( Reset_n,
                 cpu_addr, cpu_we, cpu_stb,
                 ram_stb, ram_ack,
					  rom_stb, rom_ack,
					  uart_stb, uart_ack,
					  rom_data_out, uart_data_out, ram_data_out, cpu_data_out )
begin
    case cpu_addr(10 downto 6) is
		when "00000" =>
 		  cpu_data_in <= rom_data_out;
		  rom_stb  <= cpu_stb;
		  ram_stb  <= '0';
		  uart_stb <= '0';
		when "11111" =>
		  cpu_data_in <= uart_data_out;
		  rom_stb  <= '0';
		  ram_stb  <= '0';
		  uart_stb <= cpu_stb;
		when others =>
		  cpu_data_in <= ram_data_out;
		  rom_stb  <= '0';
		  ram_stb  <= cpu_stb;
		  uart_stb <= '0';
	 end case;
 	 cpu_reset <= not Reset_n; -- CPU reset is active high
	 cpu_ack <= ram_ack or uart_ack or rom_ack;
end process;

----------------------------------------------------------------------
--
--  Processes to read and write RAM
--
----------------------------------------------------------------------

ram: process( Reset_n,
                 cpu_addr, cpu_we, cpu_stb,
                 ram_stb, ram_wr, ram_rd,
					  ram_data_out, cpu_data_out )
begin
	 ram_rd  <= Reset_n and cpu_stb and (not cpu_we);
	 ram_rdn <= not ram_rd;
    ram_wr  <= Reset_n and cpu_stb and cpu_we;
	 ram_wrn <= not ram_wr;

	 -- ************ Disable SRAM to uses RS232 *********
	 -- Note that ram_data<0> is shared with rxbit
	 -- You will have to remove all references to
	 -- the RAM to use the standard RS232 port
	 -- John Kent 1st april 2003 (no april fools)
	 --
    ram_csn <= not ram_stb;
    ram_ack <= ram_stb;

    ram_addr(16 downto 11) <= "000000";
	 ram_addr(10 downto 0) <= cpu_addr;
	 ram_data_out(7 downto 0) <= ram_data(7 downto 0);
    if ram_stb = '1' then
		if ram_wr = '1' then
		  ram_data(7 downto 0) <= cpu_data_out;
		else
        ram_data(7 downto 0) <= "ZZZZZZZZ";
		end if;
    else
      ram_data(7 downto 0)  <= "ZZZZZZZZ";
    end if;
end process;

-----------------------------------------------------------
--
-- Process to disable flash on XSA100 board
--
------------------------------------------------------------

flash: process( SysClk )
begin
    flash_csn <= '1'; -- disable flash
end process;

-----------------------------------------------------------
--
-- flash led to indicate code is working
--
-----------------------------------------------------------
increment: process (SysClk, CountL )
begin
  if Rising_Edge(SysClk) then
    countL <= countL + 1;			 
  end if;
  LED <= countL(21);
end process;

 
end my_computer; --===================== End of architecture =======================--

