--
-- Micro8 - cpu8bit2.vhd
-- Not Quite Minimal 8 Bit CPU
--
--
-- Designed By:
-- John Kent - 5 Sept 2002
-- John0457@hotmail.com
--
-- The design originally started out as a CPLD design By:
-- Tim Bscke
-- t.boescke@tuhh.de
--
-- John Kent made the following enhancements:
-- Added 8 bit index register
-- Added 4 addressing modes:
-- immediate, absolute, indexed, pc relative
-- made all instructions 2 bytes long.
-- increased addressing to 11 bits (2K bytes)
--
-- 18th May 2003
-- Changed clock edge from falling to rising for wishbone compliance
-- rename CPU control signals,
-- added hold states for Wishbone acknowledge
-- change the data output mux so it's a separate process
-- resolved some errors with the alu_ctrl not being assigned in some states
--
-- 21st May 2003
-- Hid the execute cycle in the 

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity CPU8BIT2 is
	port (	
-- Wishbone signals
     WB_CLK_I : in  std_logic;  -- clock
     WB_RST_I : in  std_logic;  -- Reset input
     WB_ADR_O : out std_logic_vector(10 downto 0); -- Adress bus          
     WB_DAT_I : in  std_logic_vector( 7 downto 0); -- DataIn Bus
     WB_DAT_O : out std_logic_vector( 7 downto 0); -- DataOut Bus
     WB_WE_O  : out std_logic;   -- Write Enable
     WB_STB_O : out std_logic;   -- Strobe
     WB_ACK_I : in  std_logic	   -- Acknowledge
);
end CPU8BIT2;

architecture CPU_ARCH of CPU8BIT2 is
	type state_type is (reset_state, fetch_state, decode_state,
                       read_abs_state, read_ind_state, read_pcr_state,
	                    write_abs_state, write_ind_state, write_pcr_state,
						     branch_state );
	type ad_type is (reset_ad, latch_ad, fetch_ad, read_ad, write_ad );
	type pc_type is (reset_pc, latch_pc, advance_pc, load_pc );
   type ea_type is (reset_ea, latch_ea, load_high_ea, load_low_ea, load_ind_ea, load_pcr_ea );
   type op_type is (reset_op, fetch_op, latch_op );
   type di_type is (reset_di, load_di, latch_di );
   type ac_type is (reset_ac, load_ac, latch_ac );
	type ix_type is (reset_ix, load_ix, latch_ix );
   type cc_type is (reset_cc, load_cc, latch_cc );
   type alu_type is (add_alu, nor_alu, tst_alu, pass_alu );

  	signal accum:       std_logic_vector(7 downto 0);
	signal xreg:        std_logic_vector(7 downto 0);
   signal cc:          std_logic_vector(2 downto 0);
	signal zflag_out:   std_logic;
	signal cflag_out:   std_logic;
	signal nflag_out:   std_logic;
	signal ea:          std_logic_vector(10 downto 0);
	signal pc:	        std_logic_vector(10 downto 0);
	signal op_code:     std_logic_vector(7 downto 3);
	signal di:          std_logic_vector(7 downto 0);
	signal state:       state_type;
	signal next_state:  state_type;
   signal ad_ctrl:     ad_type;
   signal pc_ctrl:     pc_type;
   signal ea_ctrl:     ea_type; 
   signal op_ctrl:     op_type;
	signal di_ctrl:     di_type;
	signal ac_ctrl:     ac_type;
	signal ix_ctrl:     ix_type;
	signal cc_ctrl:     cc_type;
   signal alu_ctrl:    alu_type;
   signal left:        std_logic_vector(8 downto 0);
   signal right:       std_logic_vector(8 downto 0);
	signal out_alu:     std_logic_vector(8 downto 0);
	signal strobe:      std_logic;
begin

----------------------------------
--
-- Address output multiplexer
--
----------------------------------

ad_mux: process( ad_ctrl, pc, ea )
begin
  case ad_ctrl is
    when reset_ad =>
	   WB_ADR_O <= "00000000000";
    when fetch_ad =>
	   WB_ADR_O <= pc;
	 when read_ad | write_ad =>
	   WB_ADR_O <= ea;
	 when latch_ad =>
	   WB_ADR_O <= "11111111111";
	 when others =>
	   null;
  end case;
end process;

----------------------------------
--
-- strobe output multiplexer
--
----------------------------------

strobe_mux: process( ad_ctrl )
begin
  case ad_ctrl is
    when fetch_ad | read_ad | write_ad =>
		strobe   <= '1';
	 when latch_ad | reset_ad =>
		strobe   <= '0';
	 when others =>
		null;
  end case;
  WB_STB_O <= strobe;
end process;


----------------------------------
--
-- Write output multiplexer
--
----------------------------------

write_mux: process( ad_ctrl )
begin
  case ad_ctrl is
    when write_ad =>
		WB_WE_O  <= '1';
	 when reset_ad | fetch_ad | read_ad | latch_ad =>
		WB_WE_O  <= '0';
	 when others =>
	   null;
  end case;
end process;

----------------------------------
--
-- Data output multiplexer
--
----------------------------------
da_mux : process( op_code, accum, xreg )
begin
  if op_code(7) = '0' then
      WB_DAT_O <= accum;
  else
		WB_DAT_O <= xreg;
  end if;
end process;

----------------------------------
--
-- Program Counter Control
--
----------------------------------

pc_mux: process( WB_CLK_I, WB_ACK_I, strobe, pc_ctrl, pc, ea )
variable pc_offset : std_logic_vector(10 downto 0);
begin
  case pc_ctrl is
  when advance_pc =>
	     pc_offset := "00000000001";
--  when load_pc =>
  when others =>
	     pc_offset := ea;
  end case;

  if Rising_Edge(WB_CLK_I) then
    if (strobe = '1') and (WB_ACK_I = '0') then
	   pc <= pc;
	 else
      case pc_ctrl is
	   when reset_pc =>
	     pc <= "00000000000";
	   when advance_pc | load_pc =>
	     pc <= pc + pc_offset;
	   when latch_pc =>
        pc <= pc;
	   when others =>
        null;
      end case;
	 end if;
  end if;
end process;

----------------------------------
--
-- Effective Address  Control
--
----------------------------------

ea_mux: process( WB_CLK_I, WB_ACK_I, strobe, ea_ctrl, ea, pc, xreg, WB_DAT_I )
variable ea_offset : std_logic_vector(10 downto 0);
begin
  case ea_ctrl is
  when load_ind_ea =>
	 ea_offset := ("000" & xreg);
  when others =>
--  when load_pcr_ea =>
	 ea_offset := pc;
  end case;

  if Rising_Edge(WB_CLK_I) then
    if (strobe = '1') and (WB_ACK_I = '0') then
	   ea <= ea;
	 else
      case ea_ctrl is
	   when reset_ea =>
	     ea <= "00000000000";
	   when load_high_ea =>
	     ea(10 downto 8) <= WB_DAT_I(2 downto 0);
        ea(7 downto 0)  <= ea(7 downto 0);
  	   when load_low_ea =>
	     ea(10 downto 8) <= ea(10 downto 8);
        ea(7 downto 0)  <= WB_DAT_I(7 downto 0);
	   when load_ind_ea | load_pcr_ea =>
	     ea <= ea_offset + ea;
	   when latch_ea =>
        ea <= ea;
	   when others =>
        null;
      end case;
	 end if;
  end if;
end process;

--------------------------------
--
-- Accumulator
--
--------------------------------
ac_mux : process( WB_CLK_I, WB_ACK_I, strobe, ac_ctrl, out_alu, accum )
begin
  if Rising_Edge(WB_CLK_I) then
    if (strobe = '1') and (WB_ACK_I = '0') then
	   accum <= accum;
	 else
      case ac_ctrl is
      when reset_ac =>
	     accum <= "00000000";
	   when load_ac =>
	     accum <= out_alu(7 downto 0);
	   when latch_ac =>
	     accum <= accum;
	   when others =>
	     null;
      end case;
    end if;
  end if;
end process;

--------------------------------
--
-- Index register
--
--------------------------------
ix_mux : process( WB_CLK_I, WB_ACK_I, strobe, ix_ctrl, out_alu, xreg )
begin
  if Rising_Edge(WB_CLK_I) then
    if (strobe = '1') and (WB_ACK_I = '0') then
	   xreg <= xreg;
	 else
      case ix_ctrl is
      when reset_ix =>
	     xreg <= "00000000";
	   when load_ix =>
	     xreg <= out_alu(7 downto 0);
	   when latch_ix =>
	     xreg <= xreg;
	   when others =>
		  null;
      end case;
    end if;
  end if;
end process;

----------------------------------
--
-- Condition Codes
--
----------------------------------

cc_mux: process( WB_CLK_I, WB_ACK_I, strobe, cc_ctrl, zflag_out, nflag_out, cflag_out, cc )
begin
  if Rising_Edge(WB_CLK_I) then
    if (strobe = '1') and (WB_ACK_I = '0') then
	   cc <= cc;
	 else
      case cc_ctrl is
	   when reset_cc =>
	     cc <= "000";
	   when load_cc =>
	     cc(0) <= zflag_out;
		  cc(1) <= nflag_out;
		  cc(2) <= cflag_out;
  	   when latch_cc =>
        cc <= cc;
	   when others =>
		  null;
      end case;
	 end if;
  end if;
end process;

--------------------------------
--
-- Data In
--
--------------------------------
di_mux : process( WB_CLK_I, WB_ACK_I, strobe, di_ctrl, WB_DAT_I, di )
begin
  if Rising_Edge(WB_CLK_I) then
    if (strobe = '1') and (WB_ACK_I) = '0' then
	   di <= di;
	 else
      case di_ctrl is
      when reset_di =>
	     di <= "00000000";
	   when load_di =>
	     di <= WB_DAT_I;
	   when latch_di =>
	     di <= di;
	   when others =>
		  null;
      end case;
	 end if;
  end if;
end process;


----------------------------------
--
-- op code fetch
--
----------------------------------

op_fetch: process( WB_CLK_I, WB_DAT_I, WB_ACK_I, strobe, op_code )
begin
  if Rising_Edge(WB_CLK_I) then
    if (strobe = '1') and (WB_ACK_I = '0') then
	   op_code <= op_code;
	 else
      case op_ctrl is
	   when reset_op =>
	     op_code <= "01100"; -- BRA on reset (No operation in fetch cycle)
  	   when fetch_op =>
        op_code <= WB_DAT_I(7 downto 3);
	   when latch_op =>
	     op_code <= op_code;
	   when others =>
		  null;
      end case;
	 end if;
  end if;
end process;

----------------------------------
--
-- Arithmetic Logic Unit
--
----------------------------------

alu: process( alu_ctrl, op_code, ea, accum, xreg, cc, di, left, right )
begin
    if op_code(7) = '0' then
 	   left <= "0" & accum;
	 else
	   left <= "0" & xreg;
	 end if;

    right <= "0" & di;

    case alu_ctrl is
  	 when add_alu =>
	   out_alu   <= left + right; 	-- add
	 when nor_alu =>
	   out_alu <= not( left or right );	-- nor
    when tst_alu =>
		out_alu <= left; -- sto
	 when pass_alu =>
	   out_alu <= left; -- do nothing
  	 when others =>
      null;
    end case;

    case alu_ctrl is
  	 when add_alu | nor_alu | tst_alu =>
      cflag_out <= out_alu(8);
      nflag_out <= out_alu(7);
      zflag_out <= not( out_alu(7) or out_alu(6) or out_alu(5) or out_alu(4) or
	                     out_alu(3) or out_alu(2) or out_alu(1) or out_alu(0) );
	 when pass_alu =>
		cflag_out <= cc(2);
		nflag_out <= cc(1);
		zflag_out <= cc(0);
  	 when others =>
      null;
    end case;
end process;


------------------------------------
--
-- state sequencer
--
------------------------------------

process( WB_DAT_I, accum, xreg, ea, pc, state, op_code, cc )
  begin
		   case state is
         when reset_state =>        --  released from reset
             op_ctrl    <= reset_op;
				 alu_ctrl   <= pass_alu;
             cc_ctrl    <= reset_cc;
				 ac_ctrl    <= reset_ac;
				 di_ctrl    <= reset_di;
				 ix_ctrl    <= reset_ix;
	 		    ea_ctrl    <= reset_ea;
		       pc_ctrl    <= reset_pc;
             ad_ctrl    <= reset_ad;
	 	       next_state <= fetch_state;

         when fetch_state =>        -- execute previous and fetch next instruction
			    case op_code(7 downto 5) is
			    when "000" => -- ADDA
                 alu_ctrl <= add_alu;
	              ac_ctrl  <= load_ac;
				     ix_ctrl  <= latch_ix;
                 cc_ctrl  <= load_cc;
			    when "001" => -- NORA
                 alu_ctrl <= nor_alu;
	              ac_ctrl  <= load_ac;
				     ix_ctrl  <= latch_ix;
                 cc_ctrl  <= load_cc;
			    when "010" => -- STOA
                 alu_ctrl <= tst_alu;
	              ac_ctrl  <= latch_ac;
				     ix_ctrl  <= latch_ix;
                 cc_ctrl  <= load_cc;
			    when "100" => -- ADDX
                 alu_ctrl <= add_alu;
	              ac_ctrl  <= latch_ac;
				     ix_ctrl  <= load_ix;
                 cc_ctrl  <= load_cc;
				 when "101" => -- NORX
                 alu_ctrl <= nor_alu;
	              ac_ctrl  <= latch_ac;
				     ix_ctrl  <= load_ix;
                 cc_ctrl  <= load_cc;
				 when "110" => -- STOX
                 alu_ctrl <= tst_alu;
	              ac_ctrl  <= latch_ac;
				     ix_ctrl  <= latch_ix;
                 cc_ctrl  <= load_cc;
	  			 when others => -- Bcc (No operation)
                 alu_ctrl <= pass_alu;
	              ac_ctrl  <= latch_ac;
				     ix_ctrl  <= latch_ix;
                 cc_ctrl  <= latch_cc;
  				 end case;
			    op_ctrl    <= fetch_op;
				 di_ctrl    <= latch_di;
             ea_ctrl    <= load_high_ea;
	          pc_ctrl	   <= advance_pc; 
             ad_ctrl    <= fetch_ad;
			    next_state <= decode_state;

         when decode_state =>        -- fetch address / decode instruction
             op_ctrl    <= latch_op;
				 alu_ctrl   <= pass_alu;
				 cc_ctrl    <= latch_cc;
				 ac_ctrl    <= latch_ac;
				 di_ctrl    <= load_di;
				 ix_ctrl    <= latch_ix;
             ea_ctrl    <= load_low_ea;
	  	   	 pc_ctrl    <= advance_pc;
             ad_ctrl    <= fetch_ad;
	   		 case op_code(7 downto 5) is
	          when "000" | "001" | "100" | "101" => -- ADDA #, NORA #, ADDX #, NORX #
		         case op_code(4 downto 3) is
					when "00" =>  -- Immediate
					  next_state <= fetch_state; -- execute
			      when "01" =>  -- absolute
					  next_state <= read_abs_state; -- fetch data
               when "10" => -- indexed
					  next_state <= read_ind_state; -- calc effective address
               when "11" => -- pc relative
					  next_state <= read_pcr_state; -- calc effictive address
               when others => -- should note get here
                 next_state <= fetch_state;
					end case;

				 when "010" | "110" => -- STOA or STOX
		         case op_code(4 downto 3) is
					when "00" =>
                 next_state <= fetch_state;
               when "01" => -- absolute
					  next_state <= write_abs_state;
               when "10" => -- indexed
					  next_state <= write_ind_state;
               when "11" => -- pc relative
					  next_state <= write_pcr_state;
               when others =>
                 next_state <= fetch_state;
			 		end case;

				 when "011" => -- Bcc
               case op_code(4 downto 3) is
					when "00" => -- BRA always branch
                 next_state <= branch_state;
					when "01" => -- BEQ
					  if cc(0) = '1' then
					    next_state <= branch_state;
					  else
					    next_state <= fetch_state;
					  end if;
					when "10" => -- BCS
					  if cc(2) = '1' then
					    next_state <= branch_state;
					  else
					    next_state <= fetch_state;
					  end if;
					when "11" => -- BMI
					  if cc(1) = '1' then
					    next_state <= branch_state;
					  else
					    next_state <= fetch_state;
					  end if;
					when others =>
                 next_state <= fetch_state;
					end case;

             when "111" => -- Bcc
  			      case op_code(4 downto 3) is
					when "00" => -- BRN
                 next_state <= fetch_state;
  					when "01" => -- BNE
					  if cc(0) = '0' then
					    next_state <= branch_state;
					  else
					    next_state <= fetch_state;
					  end if;
					when "10" => -- BCC
					  if cc(2) = '0' then
					    next_state <= branch_state;
					  else
                   next_state <= fetch_state;
					  end if;
					when "11" => -- BPL
					  if cc(1) = '0' then
					    next_state <= branch_state;
					  else
					    next_state <= fetch_state;
					  end if;
					when others =>
                 next_state <= fetch_state;
					end case;

	          when others => -- invalid instruction
               next_state <= fetch_state;
	    		 end case;

         when read_ind_state =>
             op_ctrl    <= latch_op;
			    alu_ctrl   <= pass_alu;
             cc_ctrl    <= latch_cc;
				 ac_ctrl    <= latch_ac;
				 di_ctrl    <= latch_di;
				 ix_ctrl    <= latch_ix;
             ea_ctrl    <= load_ind_ea;
             pc_ctrl    <= latch_pc;
             ad_ctrl    <= latch_ad;
			    next_state <= read_abs_state;

			when read_pcr_state =>  -- relative
             op_ctrl    <= latch_op;
				 alu_ctrl   <= pass_alu;
             cc_ctrl    <= latch_cc;
				 ac_ctrl    <= latch_ac;
				 di_ctrl    <= latch_di;
				 ix_ctrl    <= latch_ix;
			    ea_ctrl    <= load_pcr_ea;
			    pc_ctrl    <= latch_pc;
             ad_ctrl    <= latch_ad;
             next_state <= read_abs_state; -- fetch

			when read_abs_state => -- read data
             op_ctrl    <= latch_op;
				 alu_ctrl   <= pass_alu;
             cc_ctrl    <= latch_cc;
				 ac_ctrl    <= latch_ac;
				 di_ctrl    <= load_di;
				 ix_ctrl    <= latch_ix;
			    ea_ctrl    <= load_low_ea;
				 pc_ctrl    <= latch_pc;
             ad_ctrl    <= read_ad;
				 next_state <= fetch_state; -- test condition codes

			when write_ind_state =>  -- indexed
             op_ctrl    <= latch_op;
			    alu_ctrl   <= pass_alu;
             cc_ctrl    <= latch_cc;
				 ac_ctrl    <= latch_ac;
				 di_ctrl    <= latch_di;
				 ix_ctrl    <= latch_ix;
				 ea_ctrl    <= load_ind_ea;
			    pc_ctrl    <= latch_pc;
             ad_ctrl    <= latch_ad;
  				 next_state <= write_abs_state; -- store

  	      when write_pcr_state =>  -- relative
             op_ctrl    <= latch_op;
			    alu_ctrl   <= pass_alu;
             cc_ctrl    <= latch_cc;
				 ac_ctrl    <= latch_ac;
				 di_ctrl    <= latch_di;
				 ix_ctrl    <= latch_ix;
				 ea_ctrl    <= load_pcr_ea;
			    pc_ctrl    <= latch_pc;
             ad_ctrl    <= latch_ad;
             next_state <= write_abs_state; -- store

			when write_abs_state => -- write data
             op_ctrl    <= latch_op;
			    alu_ctrl   <= pass_alu;
             cc_ctrl    <= latch_cc;
				 ac_ctrl    <= latch_ac;
				 di_ctrl    <= latch_di;
				 ix_ctrl    <= latch_ix;
				 ea_ctrl    <= latch_ea;
				 pc_ctrl    <= latch_pc;
             ad_ctrl    <= write_ad;
				 next_state <= fetch_state; -- test condition codes

		   when branch_state => -- Bcc
             op_ctrl    <= latch_op;
			    alu_ctrl   <= pass_alu;
             cc_ctrl    <= latch_cc;
				 ac_ctrl    <= latch_ac;
				 di_ctrl    <= latch_di;
				 ix_ctrl    <= latch_ix;
				 ea_ctrl    <= latch_ea;
				 pc_ctrl    <= load_pc;
             ad_ctrl    <= latch_ad;
             next_state <= fetch_state;

			when others => -- undefine states
             null;
	  end case;						
end process;

--------------------------------
--
-- state machine
--
--------------------------------

change_state: process( WB_CLK_I, WB_RST_I, WB_ACK_I, strobe, state )
begin
  if Rising_Edge(WB_CLK_I) then
    if WB_RST_I = '1' then
 	   state <= reset_state;
    else
	   if (strobe = '1') and (WB_ACK_I = '0') then
	     state <= state;
	   else
        state <= next_state;
	   end if;
	 end if;
  end if;
end process;
-- output
	
end CPU_ARCH;
	
